/*
  Arduino.h - standard definitions for Arduino build environment
  Part of Arduino - http://www.arduino.cc/

  Copyright (c) 2007 David A. Mellis [duplicated from pins_arduino.h, not present in original]

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA

  Updated for 'xmega' core by bob frazier, S.F.T. Inc. - http://mrp3.com/

  In some cases, the xmega updates make assumptions about the pin assignments.
  See 'pins_arduino.h' for more detail.

*/


#ifndef Arduino_h
#define Arduino_h

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <avr/pgmspace.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "binary.h"

#ifdef __cplusplus
extern "C"{
#endif // __cplusplus

#define HIGH 0x1
#define LOW  0x0

#define INPUT                0x0 /* totem poll, input */
#define OUTPUT               0x1 /* totem poll, output */
#define INPUT_BUS_KEEPER     0x2 /* weak pull up/down to maintain state when switched to or in input mode */
#define INPUT_PULLUP         0x3 /* pullup resistor on input */
#define INPUT_PULLDOWN       0x4 /* pulldown resistor on input */
#define OUTPUT_OR            0x5 /* output open drain 'or', no pulldown */
#define OUTPUT_AND           0x6 /* output open drain 'and', no pullup */
#define INPUT_OR_PULLDOWN    0x7 /* output open drain 'or' with pulldown */
#define INPUT_AND_PULLUP     0x8 /* output open drain 'and' with pullup */
#define OUTPUT_OR_PULLDOWN   0x9 /* output open drain 'or' with pulldown */
#define OUTPUT_AND_PULLUP    0xa /* output open drain 'and' with pullup */
#define INPUT_OUTPUT_MASK    0xf /* mask for INPUT/OUTPUT flags */
#define INPUT_SENSE_DEFAULT  0    /* input sense default - currently 'BOTH' */
#define INPUT_SENSE_RISING   0x10 /* just rising */
#define INPUT_SENSE_FALLING  0x20 /* just falling */
#define INPUT_SENSE_BOTH     0x30 /* rising AND falling */
#define INPUT_SENSE_LEVEL    0x40 /* high level (or low if I invert it) */
#define INPUT_SENSE_DISABLED 0x50 /* buffered input disabled (most pins won't be able to use 'IN' if you do this) */
#define INPUT_SENSE_MASK     0xf0 /* mask for 'input sense' bits */
#define INPUT_OUTPUT_INVERT 0x100 /* bit for 'inverted' I/O - note that digitalRead and digitalWrite will re-invert to maintain consistency */

// NOTE:  'INPUT_OUTPUT_INVER' is primarily there to support LOW LEVEL interrupts.  if you specify this flag for normal
//        digital I/O, there will be no 'visible effect' since digitalRead and digitalWrite will "re-invert" the bit value
//        and act as if the invert flag weren't set.  That way, if you select 'LOW LEVEL' interrupt, you will read the
//        low level as a '0' (as it should be) via digitalRead, even though the value MUST be inverted for this to work.

// NOTE:  the values of 'true' and 'false' should be defined by C++ already
#define true /*0x1*/(!0) /* rather than '1' true is defined as '!0' - it's logically accurate */
#define false 0x0

#define PI 3.1415926535897932384626433832795
#define HALF_PI 1.5707963267948966192313216916398
#define TWO_PI 6.283185307179586476925286766559
#define DEG_TO_RAD 0.017453292519943295769236907684886
#define RAD_TO_DEG 57.295779513082320876798154814105

/* I do not know what these next 4 #defines do, but all 4 seem to be WRONG - bf */
#define SERIAL  0x0
#define DISPLAY 0x1

#define LSBFIRST 0
#define MSBFIRST 1

// INTERRUPT TYPE - LOW, HIGH, CHANGE, FALLING, RISING
// 'LOW' is defined as '0' already
// 'HIGH' is defined as '1' already
#define CHANGE 2
#define FALLING 3
#define RISING 4

// definitions for atmega328 etc. carried forward - not sure what this is for
#define INTERNAL 3
#define DEFAULT 1
#define EXTERNAL 0

// undefine stdlib's abs if encountered (from 'arduino' version)
#ifdef abs
#undef abs
#endif

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))
#define abs(x) ((x)>0?(x):-(x))
#define constrain(amt,low,high) ((amt)<(low)?(low):((amt)>(high)?(high):(amt)))
#define round(x)     ((x)>=0?(long)((x)+0.5):(long)((x)-0.5))
#define radians(deg) ((deg)*DEG_TO_RAD)
#define degrees(rad) ((rad)*RAD_TO_DEG)
#define sq(x) ((x)*(x))

#define interrupts() sei()
#define noInterrupts() cli()

#define clockCyclesPerMicrosecond() ( F_CPU / 1000000L )
#define clockCyclesToMicroseconds(a) ( (a) / clockCyclesPerMicrosecond() )
#define microsecondsToClockCycles(a) ( (a) * clockCyclesPerMicrosecond() )

#define lowByte(w) ((uint8_t) ((w) & 0xff))
#define highByte(w) ((uint8_t) ((w) >> 8))

#define bitRead(value, bit) (((value) >> (bit)) & 0x01)
#define bitSet(value, bit) ((value) |= (1UL << (bit)))
#define bitClear(value, bit) ((value) &= ~(1UL << (bit)))
#define bitWrite(value, bit, bitvalue) (bitvalue ? bitSet(value, bit) : bitClear(value, bit))


typedef unsigned int word;

#define bit(b) (1UL << (b))

typedef uint8_t boolean;
typedef uint8_t byte;

void init(void);

void adc_setup(void); // implemented in wiring_analog.c - configures ADC for analogRead()
// adc_setup must be called whenever exiting SLEEP MODE or ADC will malfunction
// It is automatically called from 'init()' but sleep mode typically resets the controller

void pinMode(uint8_t, uint8_t);
void digitalWrite(uint8_t, uint8_t);
int digitalRead(uint8_t);
int analogRead(uint8_t);
void analogReference(uint8_t mode); // somewhat different for xmega (default is Vcc/2)
void analogWrite(uint8_t, int);

unsigned long millis(void);
unsigned long micros(void);
void delay(unsigned long);
void delayMicroseconds(unsigned int us);
unsigned long pulseIn(uint8_t pin, uint8_t state, unsigned long timeout);

void shiftOut(uint8_t dataPin, uint8_t clockPin, uint8_t bitOrder, uint8_t val);
uint8_t shiftIn(uint8_t dataPin, uint8_t clockPin, uint8_t bitOrder);


// X M E G A   X M E G A   X M E G A   X M E G A   X M E G A   X M E G A   X M E G A   X M E G A
//
// NOTE:  for 'attachInterrupt' the 'mode' parameter indicates non-default input pins
//        and the interrupt mode, as well as the interrupt priority.  If the interrupt
//        priority is INT_MODE_PRI_DEFAULT (0), it will be assigned a 'default' value.
//        Default interrupt pin is pin 2 except for PORTR (since it doesn't have a pin 2)
//        when it has not been specified.
//
//        usage:
//          attachInterrupt(PORTD_INT0,                                     the specific interrupt vector - see pins_arduino.h
//                          my_callback,                                    user-defined callback function
//                          RISING                                          interrupt mode (can be LOW, HIGH, RISING, FALLING, CHANGE)
//                          | INT_MODE_PIN_DEFAULT                          the pin(s) to assign to this interrupt, or default pin 2 (optional)
//                          | INT_MODE_PRI_DEFAULT);                        interrupt priority, default is 'high' (optional)
//
//        Additional note, the 'pin' constants (see below) refer to the port's pin number, and
//        NOT the 'digital I/O pin' number.  See 'pins_arduino.h' for more on this.
//
// X M E G A   X M E G A   X M E G A   X M E G A   X M E G A   X M E G A   X M E G A   X M E G A

void attachInterrupt(uint8_t interruptNum, void (*)(void), int mode);
void detachInterrupt(uint8_t interruptNum);


// this next function reads data from the calibration row, including the serial # info.
// This is often referred to as the 'PRODUCT SIGNATURE ROW'.  It is xmega-specific.
uint8_t readCalibrationData(uint16_t iIndex);


// INTERRUPT MODE FLAGS - for attachInterrupt 'mode' parameter

#define INT_MODE_MODE_MASK 0x003f
#define INT_MODE_PRI_MASK  0x00c0
#define INT_MODE_PRI_DEFAULT    0
#define INT_MODE_PRI_LOW   0x0040
#define INT_MODE_PRI_MED   0x0080
#define INT_MODE_PRI_HIGH  0x00c0
#define INT_MODE_PRI_SHIFT      6 /* shift right 6 bits to get a 0, 1, 2 or 3 for priority (0 is 'default') */
#define INT_MODE_PIN_MASK  0xff00
#define INT_MODE_PIN0      0x0100
#define INT_MODE_PIN1      0x0200
#define INT_MODE_PIN2      0x0400
#define INT_MODE_PIN3      0x0800
#define INT_MODE_PIN4      0x1000
#define INT_MODE_PIN5      0x2000
#define INT_MODE_PIN6      0x4000
#define INT_MODE_PIN7      0x8000
#define INT_MODE_PIN_DEFAULT    0 /* no 'pin bits' set implies 'default' which is pin 2 on each capable port */
#define INT_MODE_PIN_SHIFT      8 /* shift right 8 bits to get the pin bits in a single byte */

// NOTE:  the 'pin' constants refer to the port's pin number, and not the digital I/O pin
//        The default 'pin 2' refers to the port's pin 2.  When DIGITAL_IO_PIN_SHIFT is
//        defined in pins_arduino.h, this will be digital I/O 0, 8, 16 for ports D, C, and E
//        respectively.  Otherwise, it will be digital I/O 2, 10, 18 for ports D, C, and E.
//        For port A it will always be digital I/O 24, and digital I/O 32 for port B.
//        See 'pins_arduino.h' for more on this.


// SETUP and LOOP (no changes from Arduino classic)

void setup(void);
void loop(void);



// On the xmega, the addresses of the port registers are
// greater than 255, so we can't store them in uint8_t's.
extern const uint16_t PROGMEM port_to_mode_PGM[];
extern const uint16_t PROGMEM port_to_input_PGM[];
extern const uint16_t PROGMEM port_to_output_PGM[];
extern const uint16_t PROGMEM digital_pin_to_control_PGM[];

// these contain index values so they CAN be uint8_t's
extern const uint8_t PROGMEM digital_pin_to_port_PGM[];
extern const uint8_t PROGMEM digital_pin_to_bit_mask_PGM[];
extern const uint8_t PROGMEM digital_pin_to_timer_PGM[];
// extern const uint8_t PROGMEM digital_pin_to_bit_PGM[];  not used on xmega

// Get the bit location within the hardware port of the given virtual pin.
// This comes from the pins_*.c file for the active board configuration.
//
// These perform slightly better as macros compared to inline functions
//
#define digitalPinToPort(P) ( pgm_read_byte( digital_pin_to_port_PGM + (P) ) )
#define digitalPinToBitMask(P) ( pgm_read_byte( digital_pin_to_bit_mask_PGM + (P) ) )
#define digitalPinToTimer(P) ( pgm_read_byte( digital_pin_to_timer_PGM + (P) ) )
#define analogInPinToBit(P) (P % 7) /* analog pin 0 = 0 (PORTA), analog pin 8 = 0 (PORTB) */
#define portOutputRegister(P) ( (volatile uint8_t *)( pgm_read_word( port_to_output_PGM + (P))) )
#define portInputRegister(P) ( (volatile uint8_t *)( pgm_read_word( port_to_input_PGM + (P))) )
#define portModeRegister(P) ( (volatile uint8_t *)( pgm_read_word( port_to_mode_PGM + (P))) )
#define pinControlRegister(P) ( (volatile uint8_t *)( pgm_read_word( digital_pin_to_control_PGM + (P))) )

#define NOT_A_PIN 0
#define NOT_A_PORT 0

#ifdef ARDUINO_MAIN
// use of '_' prefix to prevent collisions with iox64d#.h and for consistency
#define _PA 1
#define _PB 2
#define _PC 3
#define _PD 4
#define _PE 5
#define _PR 6 /* was PF */
//#define PG 7
//#define PH 8
//#define PJ 10
//#define PK 11
//#define PL 12
#endif

// modified timer definitions for xmega
// TCD2 --> TIMERD2
// TCC2 --> TIMERC2
// TCE0 --> TIMERE0
#define NOT_ON_TIMER 0
#define TIMERD2 1
#define TIMERC2 2
#define TIMERE0 3
// not using TCD0,1 nor TCC0,1
// The first 16 IO pins (PD0-PD7, PC0-PC7) will be PWM capable, as are PE0-PE3


#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#ifdef __cplusplus
#include "WCharacter.h"
#include "WString.h"
#include "HardwareSerial.h"

uint16_t makeWord(uint16_t w);
uint16_t makeWord(byte h, byte l);

#define word(...) makeWord(__VA_ARGS__)

#if 0

// these are not currently implemented - TODO implement them

unsigned long pulseIn(uint8_t pin, uint8_t state, unsigned long timeout = 1000000L);

void tone(uint8_t _pin, unsigned int frequency, unsigned long duration = 0);
void noTone(uint8_t _pin);

#endif // 0


// WMath prototypes
long random(long);
long random(long, long);
void randomSeed(unsigned int);
long map(long, long, long, long, long);

#endif // __cplusplus

#include "pins_arduino.h"

#endif // Arduino_h


