/* -*- mode: jde; c-basic-offset: 2; indent-tabs-mode: nil -*- */

/*
  Part of the Wiring project - http://wiring.uniandes.edu.co

  Copyright (c) 2004-05 Hernando Barragan

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA

  Modified 24 November 2006 by David A. Mellis
  Modified 1 August 2010 by Mark Sproul

  Updated for 'xmega' core by bob frazier, S.F.T. Inc. - http://mrp3.com/

  In some cases, the xmega updates make assumptions about the pin assignments.
  See 'pins_arduino.h' for more detail.

*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>

#include "wiring_private.h"

// interrupts on the xmega are handled differently than the mega
// there are 2 interrupt vectors for each port.  Typical xmega
// will use ports A, B, C, D, E, and R.  The vectors for those
// are as follows:
//
//   PORTn_INT0_vect
//   - and -
//   PORTn_INT1_vect
//
// where 'n' is A, B, C, D, E, or R
//
// Additional vectors are:
//
//   OSC_XOSCF_vect  (external oscillator failure, NMI)
//
//   RTC_OVF_vect    (real-time clock overflow)
//   RTC_COMP_vect   (real-time clock compare)
//
//   TWIC_TWIS_vect  (2-wire slave on port C)
//   TWIC_TWIM_vect  (2-wire master on port C)
//   TWIE_TWIS_vect  (2-wire slave on port E)
//   TWIE_TWIM_vect  (2-wire master on port E)
//
//   timers - 'n' is C or D
//   TCn0_OVF_vect   (n timer 0 overflow)
//   TCn1_OVF_vect   (n timer 1 overflow)
//   TCn2_LUNF_vect  (n timer 2 low byte underflow)
//   TCn2_HUNF_vect  (n timer 2 high byte underflow)
//   TCE0_OVF_vect   (E timer 0 overflow)
//   TCn0_ERR_vect   (n timer 0 error)
//   TCn1_ERR_vect   (n timer 1 error)
//   TCE0_ERR_vect   (E timer 0 error)
//   TCn0_CCA_vect   (n timer 0 compare or capture A)
//   TCn1_CCA_vect   (n timer 1 compare or capture A)
//   TCn2_LCMPA_vect (n timer 2 low-byte compare or capture A)
//   TCE0_CCA_vect   (E timer 0 compare or capture A)
//   TCn0_CCB_vect   (n timer 0 compare or capture B)
//   TCn1_CCB_vect   (n timer 1 compare or capture B)
//   TCn2_LCMPB_vect (n timer 2 low-byte compare or capture B)
//   TCE0_CCB_vect   (E timer 0 compare or capture B)
//   TCn0_CCC_vect   (n timer 0 compare or capture C)
//   TCn2_LCMPC_vect (n timer 2 low-byte compare or capture C)
//   TCE0_CCC_vect   (E timer 0 compare or capture C)
//   TCn0_CCD_vect   (n timer 0 compare or capture D)
//   TCn2_LCMPD_vect (n timer 2 low-byte compare or capture D)
//   TCE0_CCD_vect   (E timer 0 compare or capture D)
//
//   SPIn_INT_vect   (SPI C or D)
//
//   USARTn0_RXC_vect (USART 'n' [C or D] receive complete)
//   USARTn0_DRE_vect (USART 'n' [C or D] data reg empty)
//   USARTn0_TXC_vect (USART 'n' [C or D] transmit complete)
//
// NOTE:  a 'USARTE' interrupt vector also exists, but isn't
// implemented on the D4 series
//
// ASYNC interrupts are only possible on pin 2 for each of the 5
// ports ('R' only has 2 pins, 0 and 1, so no async interrupt).
// Sleep modes typically need ASYNC interrupts to wake up.

// The interrupt will be handled for a particular port, and not for
// a particular interrupt.  The 'attachInterrupt' function will default
// to pin 2 (asynchronous interrupt) unless otherwise specified in the
// 'mode' parameter.

// BOOTLOADER NOTES
//
// Bit 6 of the CTRL reg must be assigned to '0'.  Bit 7 can be assigned
// to '1' to enable 'round robin' scheduling using the priority bits.
// Bits 0-2 (HILVLEN, MEDLVLEN, LOLVLEN) should also be assigned to '1'
// to allow all 3 interrupt levels to execute.  ('D' manual 10.8.3 pg 102)
// The CTRL reg can be assigned to b10000111 to accomplish this.  Bit 6
// needs to use the "configuration change protection" method to change it
// and may need to be assigned separately.



// interrupt mode - predefined values are LOW, CHANGE, RISING, FALLING, HIGH
// additional bits are 'or'd with mode

static volatile voidFuncPtr intFunc[EXTERNAL_NUM_INTERRUPTS];
static volatile uint8_t intPins[EXTERNAL_NUM_INTERRUPTS]; // added - store pins for this interrupt
// volatile static voidFuncPtr twiIntFunc;


#define NEW_INT_CODE /* use the smaller, less 'brute force' versions */

// NOTE: I _HATE_ K&R style so I'll make it Allman style as I go along...

#ifdef NEW_INT_CODE

void attachInterrupt(uint8_t interruptNum, void (*userFunc)(void), int mode)
{
uint8_t iPinBits, iPriBits, iModeBits, iInv, iNum, iMask;
uint8_t oldSREG;
PORT_t *port;


  if(interruptNum >= EXTERNAL_NUM_INTERRUPTS)
  {
    return;
  }

  iPinBits = (uint8_t)((mode & INT_MODE_PIN_MASK) >> 8);

  if(!iPinBits)
  {
    if(interruptNum == PORTR_INT0 || interruptNum == PORTR_INT1) // not valid for these
    {
      return; // do nothing (for now)
    }

    iPinBits = _BV(2); // set bit for pin 2 if none specified [i.e. 'default']
  }

  iPriBits = (mode & INT_MODE_PRI_MASK)
           >> INT_MODE_PRI_SHIFT;

  if(!iPriBits) // not assigned
  {
    iPriBits = 3; // for now, just use the highest priority
  }

  mode &= INT_MODE_MODE_MASK;
  iInv = 0;

  if(mode == LOW) // normally will be this, for backward hardware compatibility
  {
    iModeBits = PORT_ISC_LEVEL_gc; // b011, high level continuously generates events
  }
  else if(mode == HIGH) // these constants correspond to the mega's bit mask on ISC00,ISC10
  {
    iModeBits = PORT_ISC_LEVEL_gc; // b011, high level continuously generates events
    iInv = 1;                      // invert input (so 'high level' becomes 'low level')

    // NOTE:  this was verified by experimentation.  The documentation is misleading, suggesting
    //        that a LEVEL interrupt triggered on HIGH, not on LOW.  But it triggers on LOW.  So
    //        if you want HIGH, you must invert it.  Not the other way around.  Yeah.
  }
  else if(mode == CHANGE)
  {
    iModeBits = PORT_ISC_BOTHEDGES_gc; // BOTHEDGES - see table 11-6
  }
  else if(mode == RISING)
  {
    iModeBits = PORT_ISC_RISING_gc; // b001, RISING
  }
  else if(mode == FALLING)
  {
    iModeBits = PORT_ISC_FALLING_gc; // b010, FALLING
  }
  else
  {
    iModeBits = PORT_ISC_BOTHEDGES_gc; // BOTH (the default - note INTPUT_DISABLED (sic) won't buffer the input, so it's useless except for analog channels)
  }

  if(iInv)
  {
    iModeBits |= _BV(PORT_INVEN_bp); // set the 'inverted' bit
  }

  oldSREG = SREG; // store the interrupt flag basically

  cli(); // disable interrupts for a bit

  intFunc[interruptNum] = userFunc;
  intPins[interruptNum] = iPinBits; // save what pins I used

  // Enable the interrupt (smaller code to use if/else and pointer)

  iNum = 0;
  if(interruptNum == PORTA_INT0 || interruptNum == PORTA_INT1)
  {
    port = &PORTA;
    if(interruptNum == PORTA_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTB_INT0 || interruptNum == PORTB_INT1)
  {
    port = &PORTB;
    if(interruptNum == PORTB_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTC_INT0 || interruptNum == PORTC_INT1)
  {
    port = &PORTC;
    if(interruptNum == PORTC_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTD_INT0 || interruptNum == PORTD_INT1)
  {
    port = &PORTD;
    if(interruptNum == PORTD_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTE_INT0 || interruptNum == PORTE_INT1)
  {
    port = &PORTE;
    if(interruptNum == PORTE_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTR_INT0 || interruptNum == PORTR_INT1)
  {
    port = &PORTR;
    if(interruptNum == PORTR_INT1)
    {
      iNum = 1;
    }
  }
  else
  {
    return; // do nothing
  }

  if(!iNum)
  {
    // set interrupt mask for PORT A, int 0 vector
    port->INT0MASK |= iPinBits; // enable int zero for these pins
    port->INTCTRL = (port->INTCTRL & ~(PORT_INT0LVL_gm))
                  | (iPriBits & 3);
  }
  else // if(iNum == 1)
  {
    port->INT1MASK |= iPinBits; // enable int zero for these pins
    port->INTCTRL = (port->INTCTRL & ~(PORT_INT1LVL_gm))
                  | ((iPriBits & 3) << PORT_INT1LVL_gp);
  }

  for(iNum=0, iMask = 1; iNum < 8; iNum++, iMask <<= 1)
  {
    register8_t *pCTRL = &(port->PIN0CTRL) + iNum; // treat PIN0CTRL through PIN7CTRL as an array

    // set corresponding 'type' in the interrupt control regs for the individual bits

    if(iPinBits & iMask) // is this bit set in 'iPinBits'?
    {
      // enable the interrupt pin's mode bits and assign the 'invert' flag as needed

      *pCTRL = (*pCTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
             | iModeBits;
    }
  }

  SREG = oldSREG; // restore it, interrupts (probably) re-enabled
  // NOTE that this may throw an interrupt right away
}

#else // NEW_INT_CODE

void attachInterrupt(uint8_t interruptNum, void (*userFunc)(void), int mode)
{
uint8_t iPinBits, iPriBits, iModeBits, iInv;
uint8_t oldSREG;


  if(interruptNum >= EXTERNAL_NUM_INTERRUPTS)
  {
    return;
  }

  iPinBits = (uint8_t)((mode & INT_MODE_PIN_MASK) >> 8);

  if(!iPinBits)
  {
    if(interruptNum == PORTR_INT0 || interruptNum == PORTR_INT1) // not valid for these
    {
      return; // do nothing (for now)
    }

    iPinBits = _BV(2); // set bit for pin 2 if none specified [i.e. 'default']
  }

  iPriBits = (mode & INT_MODE_PRI_MASK)
           >> INT_MODE_PRI_SHIFT;

  if(!iPriBits) // not assigned
  {
    iPriBits = 3; // for now, just use the highest priority
  }

  mode &= INT_MODE_MODE_MASK;
  iInv = 0;

  if(mode == LOW) // normally will be this, for backward hardware compatibility
  {
    iModeBits = PORT_ISC_LEVEL_gc; // b011, high level continuously generates events
  }
  else if(mode == HIGH) // these constants correspond to the mega's bit mask on ISC00,ISC10
  {
    iModeBits = PORT_ISC_LEVEL_gc; // b011, high level continuously generates events
    iInv = 1;                      // invert input (so 'high level' becomes 'low level')

    // NOTE:  this was verified by experimentation.  The documentation is misleading, suggesting
    //        that a LEVEL interrupt triggered on HIGH, not on LOW.  But it triggers on LOW.  So
    //        if you want HIGH, you must invert it.  Not the other way around.  Yeah.
  }
  else if(mode == CHANGE)
  {
    iModeBits = PORT_ISC_BOTHEDGES_gc; // BOTHEDGES - see table 11-6
  }
  else if(mode == RISING)
  {
    iModeBits = PORT_ISC_RISING_gc; // b001, RISING
  }
  else if(mode == FALLING)
  {
    iModeBits = PORT_ISC_FALLING_gc; // b010, FALLING
  }
  else
  {
    iModeBits = PORT_ISC_BOTHEDGES_gc; // BOTH (the default - note INTPUT_DISABLED (sic) won't buffer the input, so it's useless except for analog channels)
  }

  if(iInv)
  {
    iModeBits |= _BV(PORT_INVEN_bp); // set the 'inverted' bit
  }

  oldSREG = SREG; // store the interrupt flag basically

  cli(); // disable interrupts for a bit

  intFunc[interruptNum] = userFunc;
  intPins[interruptNum] = iPinBits; // save what pins I used


  // Enable the interrupt (smaller code to use if/else)

  if(interruptNum == PORTA_INT0 || interruptNum == PORTA_INT1)
  {
    if(interruptNum == PORTA_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTA_INT0MASK |= iPinBits; // enable int zero for these pins
      PORTA_INTCTRL = (PORTA_INTCTRL & ~(PORT_INT0LVL_gm))
                    | (iPriBits & 3);
    }
    else if(interruptNum == PORTA_INT1)
    {
      // set interrupt mask for PORT A, int 1 vector
      PORTA_INT1MASK |= iPinBits; // enable int one for these pins
      PORTA_INTCTRL = (PORTA_INTCTRL & ~(PORT_INT1LVL_gm))
                    | ((iPriBits & 3) << PORT_INT1LVL_gp);
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?
    //        also possible, use 'pinMode()' with appropriate flags...

    if(iPinBits & _BV(0))
    {
      PORTA_PIN0CTRL = (PORTA_PIN0CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(1))
    {
      PORTA_PIN1CTRL = (PORTA_PIN1CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(2))
    {
      PORTA_PIN2CTRL = (PORTA_PIN2CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(3))
    {
      PORTA_PIN3CTRL = (PORTA_PIN3CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(4))
    {
      PORTA_PIN4CTRL = (PORTA_PIN4CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(5))
    {
      PORTA_PIN5CTRL = (PORTA_PIN5CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(6))
    {
      PORTA_PIN6CTRL = (PORTA_PIN6CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(7))
    {
      PORTA_PIN7CTRL = (PORTA_PIN7CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
  }
  else if(interruptNum == PORTB_INT0 || interruptNum == PORTB_INT1)
  {
    if(interruptNum == PORTB_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTB_INT0MASK |= iPinBits & 0xf; // enable int zero for these pins
      PORTB_INTCTRL = (PORTB_INTCTRL & ~(PORT_INT0LVL_gm))
                    | (iPriBits & 3);
    }
    else if(interruptNum == PORTB_INT1)
    {
      // set interrupt mask for PORT A, int 1 vector
      PORTB_INT1MASK |= iPinBits & 0xf; // enable int one for these pins
      PORTB_INTCTRL = (PORTB_INTCTRL & ~(PORT_INT1LVL_gm))
                    | ((iPriBits & 3) << PORT_INT1LVL_gp);
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTB_PIN0CTRL = (PORTB_PIN0CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(1))
    {
      PORTB_PIN1CTRL = (PORTB_PIN1CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(2))
    {
      PORTB_PIN2CTRL = (PORTB_PIN2CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(3))
    {
      PORTB_PIN3CTRL = (PORTB_PIN3CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
  }
  else if(interruptNum == PORTC_INT0 || interruptNum == PORTC_INT1)
  {
    if(interruptNum == PORTC_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTC_INT0MASK |= iPinBits; // enable int zero for these pins
      PORTC_INTCTRL = (PORTC_INTCTRL & ~(PORT_INT0LVL_gm))
                    | (iPriBits & 3);
    }
    else if(interruptNum == PORTC_INT1)
    {
      // set interrupt mask for PORT A, int 1 vector
      PORTC_INT1MASK |= iPinBits; // enable int one for these pins
      PORTC_INTCTRL = (PORTC_INTCTRL & ~(PORT_INT1LVL_gm))
                    | ((iPriBits & 3) << PORT_INT1LVL_gp);
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTC_PIN0CTRL = (PORTC_PIN0CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(1))
    {
      PORTC_PIN1CTRL = (PORTC_PIN1CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(2))
    {
      PORTC_PIN2CTRL = (PORTC_PIN2CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(3))
    {
      PORTC_PIN3CTRL = (PORTC_PIN3CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(4))
    {
      PORTC_PIN4CTRL = (PORTC_PIN4CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(5))
    {
      PORTC_PIN5CTRL = (PORTC_PIN5CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(6))
    {
      PORTC_PIN6CTRL = (PORTC_PIN6CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(7))
    {
      PORTC_PIN7CTRL = (PORTC_PIN7CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
  }
  else if(interruptNum == PORTD_INT0 || interruptNum == PORTD_INT1)
  {
    if(interruptNum == PORTD_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTD_INT0MASK |= iPinBits; // enable int zero for these pins
      PORTD_INTCTRL = (PORTD_INTCTRL & ~(PORT_INT0LVL_gm))
                    | (iPriBits & 3);
    }
    else if(interruptNum == PORTD_INT1)
    {
      // set interrupt mask for PORT A, int 1 vector
      PORTD_INT1MASK |= iPinBits; // enable int one for these pins
      PORTD_INTCTRL = (PORTD_INTCTRL & ~(PORT_INT1LVL_gm))
                    | ((iPriBits & 3) << PORT_INT1LVL_gp);
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTD_PIN0CTRL = (PORTD_PIN0CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(1))
    {
      PORTD_PIN1CTRL = (PORTD_PIN1CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(2))
    {
      PORTD_PIN2CTRL = (PORTD_PIN2CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(3))
    {
      PORTD_PIN3CTRL = (PORTD_PIN3CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(4))
    {
      PORTD_PIN4CTRL = (PORTD_PIN4CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(5))
    {
      PORTD_PIN5CTRL = (PORTD_PIN5CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(6))
    {
      PORTD_PIN6CTRL = (PORTD_PIN6CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(7))
    {
      PORTD_PIN7CTRL = (PORTD_PIN7CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
  }
  else if(interruptNum == PORTE_INT0 || interruptNum == PORTE_INT1)
  {
    if(interruptNum == PORTE_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTE_INT0MASK |= iPinBits & 0xf; // enable int zero for these pins
      PORTE_INTCTRL = (PORTE_INTCTRL & ~(PORT_INT0LVL_gm))
                    | (iPriBits & 3);
    }
    else if(interruptNum == PORTE_INT1)
    {
      // set interrupt mask for PORT A, int 1 vector
      PORTE_INT1MASK |= iPinBits & 0xf; // enable int one for these pins
      PORTE_INTCTRL = (PORTE_INTCTRL & ~(PORT_INT1LVL_gm))
                    | ((iPriBits & 3) << PORT_INT1LVL_gp);
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTE_PIN0CTRL = (PORTE_PIN0CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(1))
    {
      PORTE_PIN1CTRL = (PORTE_PIN1CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(2))
    {
      PORTE_PIN2CTRL = (PORTE_PIN2CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(3))
    {
      PORTE_PIN3CTRL = (PORTE_PIN3CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
  }
  else if(interruptNum == PORTR_INT0 || interruptNum == PORTR_INT1)
  {
    if(interruptNum == PORTR_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTR_INT0MASK |= iPinBits & 0x3; // enable int zero for these pins
      PORTR_INTCTRL = (PORTR_INTCTRL & ~(PORT_INT0LVL_gm))
                    | (iPriBits & 3);
    }
    else if(interruptNum == PORTR_INT1)
    {
      // set interrupt mask for PORT A, int 1 vector
      PORTR_INT1MASK |= iPinBits & 0x3; // enable int one for these pins
      PORTR_INTCTRL = (PORTR_INTCTRL & ~(PORT_INT1LVL_gm))
                    | ((iPriBits & 3) << PORT_INT1LVL_gp);
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTR_PIN0CTRL = (PORTR_PIN0CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
    if(iPinBits & _BV(1))
    {
      PORTR_PIN1CTRL = (PORTR_PIN1CTRL & ~(PORT_ISC_gm | PORT_INVEN_bm))
                     | iModeBits;
    }
  }

  SREG = oldSREG; // restore it, interrupts (probably) re-enabled

}
#endif // NEW_INT_CODE


#ifdef NEW_INT_CODE
void detachInterrupt(uint8_t interruptNum)
{
uint8_t iPinBits, iNum, iMask;
uint8_t oldSREG;
PORT_t *port;


  // NOTE:  this function will turn OFF the 'invert' bit if it's set for a HIGH level interrupt
  //        and digitalRead _SHOULD_ be consistent before/after this call.

  if(interruptNum >= EXTERNAL_NUM_INTERRUPTS)
  {
    return;
  }

  oldSREG = SREG; // keep track of interrupt flag state

  cli(); // clear the interrupt flag

  // grab 'pin bits' so I know what to flip around
  iPinBits = intPins[interruptNum]; // what I used when I added it

  intFunc[interruptNum] = 0;
  intPins[interruptNum] = 0; // zero both of these

  // disable the interrupt

  // Enable the interrupt (smaller code to use if/else and pointer)

  iNum = 0;
  if(interruptNum == PORTA_INT0 || interruptNum == PORTA_INT1)
  {
    port = &PORTA;
    if(interruptNum == PORTA_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTB_INT0 || interruptNum == PORTB_INT1)
  {
    port = &PORTB;
    if(interruptNum == PORTB_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTC_INT0 || interruptNum == PORTC_INT1)
  {
    port = &PORTC;
    if(interruptNum == PORTC_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTD_INT0 || interruptNum == PORTD_INT1)
  {
    port = &PORTD;
    if(interruptNum == PORTD_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTE_INT0 || interruptNum == PORTE_INT1)
  {
    port = &PORTE;
    if(interruptNum == PORTE_INT1)
    {
      iNum = 1;
    }
  }
  else if(interruptNum == PORTR_INT0 || interruptNum == PORTR_INT1)
  {
    port = &PORTR;
    if(interruptNum == PORTR_INT1)
    {
      iNum = 1;
    }
  }
  else
  {
    return; // do nothing
  }

  if(!iNum)
  {
    // set interrupt mask for PORT A, int 0 vector
    port->INT0MASK = 0;                  // disable interrupts - TODO, read this instead of 'iPinBits' ?
    port->INTCTRL &= ~(PORT_INT0LVL_gm); // set interrupt control to 'OFF'
    port->INTFLAGS = _BV(0);             // clear the int 0 flag
  }
  else // if(iNum == 1)
  {
    port->INT1MASK = 0;                  // disable interrupts - TODO, read this instead of 'iPinBits' ?
    port->INTCTRL &= ~(PORT_INT1LVL_gm); // set interrupt control to 'OFF'
    port->INTFLAGS = _BV(1);             // clear the int 1 flag
  }

  for(iNum=0, iMask = 1; iNum < 8; iNum++, iMask <<= 1)
  {
    register8_t *pCTRL = &(port->PIN0CTRL) + iNum; // treat PIN0CTRL through PIN7CTRL as an array

    // set corresponding 'type' in the interrupt control regs for the individual bits

    if(iPinBits & iMask) // is this bit set in 'iPinBits'?
    {
      *pCTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // turn off invert flag and reset to 'BOTH' (the default)
    }
  }

  SREG = oldSREG; // restore it, interrupts (probably) re-enabled

}

#else // NEW_INT_CODE

void detachInterrupt(uint8_t interruptNum)
{
uint8_t iPinBits;
uint8_t oldSREG;

  // NOTE:  this function will turn OFF the 'invert' bit if it's set for a HIGH level interrupt
  //        and digitalRead _SHOULD_ be consistent before/after this call.

  if(interruptNum >= EXTERNAL_NUM_INTERRUPTS)
  {
    return;
  }

  oldSREG = SREG; // keep track of interrupt flag state

  cli(); // clear the interrupt flag

  // grab 'pin bits' so I know what to flip around
  iPinBits = intPins[interruptNum]; // what I used when I added it

  intFunc[interruptNum] = 0;
  intPins[interruptNum] = 0; // zero both of these

  // disable the interrupt

  // Enable the interrupt (smaller code to use if/else)

  if(interruptNum == PORTA_INT0 || interruptNum == PORTA_INT1)
  {
    if(interruptNum == PORTA_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTA_INT0MASK = 0;                  // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTA_INTCTRL &= ~(PORT_INT0LVL_gm); // set interrupt control to 'OFF'
      PORTA_INTFLAGS = _BV(0);             // clear the int 0 flag
    }
    else if(interruptNum == PORTA_INT1)
    {
      PORTA_INT1MASK = 0;                  // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTA_INTCTRL &= ~(PORT_INT1LVL_gm); // set interrupt control to 'OFF'
      PORTA_INTFLAGS = _BV(1);             // clear the int 1 flag
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTA_PIN0CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(1))
    {
      PORTA_PIN1CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(2))
    {
      PORTA_PIN2CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(3))
    {
      PORTA_PIN3CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(4))
    {
      PORTA_PIN4CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(5))
    {
      PORTA_PIN5CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(6))
    {
      PORTA_PIN6CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(7))
    {
      PORTA_PIN7CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
  }
  else if(interruptNum == PORTB_INT0 || interruptNum == PORTB_INT1)
  {
    if(interruptNum == PORTB_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTB_INT0MASK = 0;                    // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTB_INTCTRL &= ~(PORT_INT0LVL_gm);   // set interrupt control to 'OFF'
      PORTB_INTFLAGS = _BV(0);               // clear the int 0 flag
    }
    else if(interruptNum == PORTB_INT1)
    {
      PORTB_INT1MASK = 0;                   // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTB_INTCTRL &= ~(PORT_INT1LVL_gm);  // set interrupt control to 'OFF'
      PORTB_INTFLAGS = _BV(1);               // clear the int 1 flag
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTB_PIN0CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(1))
    {
      PORTB_PIN1CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(2))
    {
      PORTB_PIN2CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(3))
    {
      PORTB_PIN3CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
  }
  else if(interruptNum == PORTC_INT0 || interruptNum == PORTC_INT1)
  {
    if(interruptNum == PORTC_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTC_INT0MASK = 0;                   // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTC_INTCTRL &= ~(PORT_INT0LVL_gm);  // set interrupt control to 'OFF'
      PORTC_INTFLAGS = _BV(0);              // clear the int 0 flag
    }
    else if(interruptNum == PORTC_INT1)
    {
      PORTC_INT1MASK = 0;                   // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTC_INTCTRL &= ~(PORT_INT1LVL_gm);  // set interrupt control to 'OFF'
      PORTC_INTFLAGS = _BV(1);              // clear the int 1 flag
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTC_PIN0CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(1))
    {
      PORTC_PIN1CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(2))
    {
      PORTC_PIN2CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(3))
    {
      PORTC_PIN3CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(4))
    {
      PORTC_PIN4CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(5))
    {
      PORTC_PIN5CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(6))
    {
      PORTC_PIN6CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(7))
    {
      PORTC_PIN7CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
  }
  else if(interruptNum == PORTD_INT0 || interruptNum == PORTD_INT1)
  {
    if(interruptNum == PORTD_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTD_INT0MASK = 0;                   // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTD_INTCTRL &= ~(PORT_INT0LVL_gm);  // set interrupt control to 'OFF'
      PORTD_INTFLAGS = _BV(0);              // clear the int 0 flag
    }
    else if(interruptNum == PORTD_INT1)
    {
      PORTD_INT1MASK = 0;                   // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTD_INTCTRL &= ~(PORT_INT1LVL_gm);  // set interrupt control to 'OFF'
      PORTD_INTFLAGS = _BV(1);              // clear the int 1 flag
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTD_PIN0CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(1))
    {
      PORTD_PIN1CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(2))
    {
      PORTD_PIN2CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(3))
    {
      PORTD_PIN3CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(4))
    {
      PORTD_PIN4CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(5))
    {
      PORTD_PIN5CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(6))
    {
      PORTD_PIN6CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(7))
    {
      PORTD_PIN7CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
  }
  else if(interruptNum == PORTE_INT0 || interruptNum == PORTE_INT1)
  {
    if(interruptNum == PORTE_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTE_INT0MASK = 0;                   // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTE_INTCTRL &= ~(PORT_INT0LVL_gm);  // set interrupt control to 'OFF'
      PORTE_INTFLAGS = _BV(0);              // clear the int 0 flag
    }
    else if(interruptNum == PORTE_INT1)
    {
      PORTE_INT1MASK = 0;                   // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTE_INTCTRL &= ~(PORT_INT1LVL_gm);  // set interrupt control to 'OFF'
      PORTE_INTFLAGS = _BV(1);              // clear the int 1 flag
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    if(iPinBits & _BV(0))
    {
      PORTE_PIN0CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(1))
    {
      PORTE_PIN1CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(2))
    {
      PORTE_PIN2CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
    if(iPinBits & _BV(3))
    {
      PORTE_PIN3CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTH' (default)
    }
  }
  else if(interruptNum == PORTR_INT0 || interruptNum == PORTR_INT1)
  {
    if(interruptNum == PORTR_INT0)
    {
      // set interrupt mask for PORT A, int 0 vector
      PORTR_INT0MASK = 0;                   // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTR_INTCTRL &= ~(PORT_INT0LVL_gm);  // set interrupt control to 'OFF'
      PORTR_INTFLAGS = _BV(0);              // clear the int 0 flag
    }
    else if(interruptNum == PORTR_INT1)
    {
      PORTR_INT1MASK = 0;                   // disable interrupts - TODO, read this instead of 'iPinBits' ?
      PORTR_INTCTRL &= ~(PORT_INT1LVL_gm);  // set interrupt control to 'OFF'
      PORTR_INTFLAGS = _BV(1);              // clear the int 1 flag
    }

    // set corresponding 'type' in the interrupt control regs for the individual bits
    // TODO:  find a BETTER way, maybe use 'PORT_t' structure?

    // NOTE:  PORT 'R' does not support 'INTPUT_DISABLE' (sic) - D manual table 11-6 pg 118
    if(iPinBits & _BV(0))
    {
      PORTR_PIN0CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTHEDGES'
    }
    if(iPinBits & _BV(1))
    {
      PORTR_PIN1CTRL &= ~(PORT_ISC_gm | PORT_INVEN_bm); // set to 'BOTHEDGES'
    }
  }

  SREG = oldSREG; // restore it, interrupts (probably) re-enabled

}
#endif // NEW_INT_CODE

ISR(PORTA_INT0_vect)
{
  if(intFunc[PORTA_INT0])
    intFunc[PORTA_INT0]();
}

ISR(PORTA_INT1_vect)
{
  if(intFunc[PORTA_INT1])
    intFunc[PORTA_INT1]();
}

ISR(PORTB_INT0_vect)
{
  if(intFunc[PORTB_INT0])
    intFunc[PORTB_INT0]();
}

ISR(PORTB_INT1_vect)
{
  if(intFunc[PORTB_INT1])
    intFunc[PORTB_INT1]();
}

ISR(PORTC_INT0_vect)
{
  if(intFunc[PORTC_INT0])
    intFunc[PORTC_INT0]();
}

ISR(PORTC_INT1_vect)
{
  if(intFunc[PORTC_INT1])
    intFunc[PORTC_INT1]();
}

ISR(PORTD_INT0_vect)
{
  if(intFunc[PORTD_INT0])
    intFunc[PORTD_INT0]();
}

ISR(PORTD_INT1_vect)
{
  if(intFunc[PORTD_INT1])
    intFunc[PORTD_INT1]();
}

ISR(PORTE_INT0_vect)
{
  if(intFunc[PORTE_INT0])
    intFunc[PORTE_INT0]();
}

ISR(PORTE_INT1_vect)
{
  if(intFunc[PORTE_INT1])
    intFunc[PORTE_INT1]();
}

ISR(PORTR_INT0_vect)
{
  if(intFunc[PORTR_INT0])
    intFunc[PORTR_INT0]();
}

ISR(PORTR_INT1_vect)
{
  if(intFunc[PORTR_INT1])
    intFunc[PORTR_INT1]();
}



